#!/usr/bin/python
#
# calibcant - tools for thermally calibrating AFM cantilevers
#
# Copyright (C) 2007,2008, William Trevor King
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# The author may be contacted at <wking@drexel.edu> on the Internet, or
# write to Trevor King, Drexel University, Physics Dept., 3141 Chestnut St.,
# Philadelphia PA 19104, USA.

"""
Aquire, save, and load cantilever calibration bump data.
For measuring photodiode sensitivity.

W. Trevor King Dec. 2007 - Oct. 2008

The relevent physical quantities are :
 Vzp_out  Output z-piezo voltage (what we generate)
 Vzp      Applied z-piezo voltage (after external ZPGAIN)
 Zp       The z-piezo position
 Zcant    The cantilever vertical deflection
 Vphoto   The photodiode vertical deflection voltage (what we measure)

Which are related by the parameters :
 zpGain           Vzp_out / Vzp
 zpSensitivity    Zp / Vzp
 photoSensitivity Vphoto / Zcant

Cantilever calibration assumes a pre-calibrated z-piezo (zpSensitivity) and
amplifier (zpGain).  In our lab, the z-piezo is calibrated by imaging a
calibration sample, which has features with well defined sizes, and the gain
is set with a knob on the Nanoscope.

photoSensitivity is measured by bumping the cantilever against the surface,
where Zp = Zcant (see the bump_*() family of functions)
The measured slope Vphoto/Vout is converted to photoSensitivity via
Vphoto/Vzp_out * Vzp_out/Vzp  * Vzp/Zp   *    Zp/Zcant =    Vphoto/Zcant
 (measured)      (1/zpGain) (1/zpSensitivity)    (1)  (photoSensitivity)

We do all these measurements a few times to estimate statistical errors.

The functions are layed out in the families:
 bump_*()
For each family, * can be any of :
 aquire       get real-world data
 save         store real-world data to disk
 load         get real-world data from disk
 analyze      interperate the real-world data.
 plot         show a nice graphic to convince people we're working :p
 load_analyze_tweaked
              read a file with a list of paths to previously saved real world data
              load each file using *_load(), analyze using *_analyze(), and
              optionally plot using *_plot().
              Intended for re-processing old data.
A family name without any _* extension (e.g. bump()),
 runs *_aquire(), *_save(), *_analyze(), *_plot().
"""

import numpy
import time 
import data_logger
import z_piezo_utils
import FFT_tools
import linfit
from calibcant_bump_analyze import bump_analyze

LOG_DATA = True  # quietly grab all real-world data and log to LOG_DIR
LOG_DIR = '$DEFAULT$/calibrate_cantilever'

TEXT_VERBOSE = True      # for debugging


# bump family

def bump_aquire(zpiezo, push_depth, npoints, freq) :
    """
    Ramps closer push_depth and returns to the original position.
    Inputs:
     zpiezo     an opened zpiezo.zpiezo instance
     push_depth distance to approach, in nm
     npoints    number points during the approach and during the retreat
     freq       rate at which data is aquired
     log_dir    directory to log data to (see data_logger.py).
                None to turn off logging (see also the global LOG_DATA).
    Returns the aquired ramp data dictionary, with data in DAC/ADC bits.
    """
    # generate the bump output
    start_pos = zpiezo.curPos()
    pos_dist = zpiezo.pos_nm2out(push_depth) - zpiezo.pos_nm2out(0)
    close_pos = start_pos + pos_dist
    appr = linspace(start_pos, close_pos, npoints)
    retr = linspace(close_pos, start_pos, npoints)
    out = concatenate((appr, retr))
    # run the bump, and measure deflection
    if TEXT_VERBOSE :
        print "Bump %g nm" % push_depth
    data = zpiezo.ramp(out, freq)
    # default saving, so we have a log in-case the operator is lazy ;)
    if LOG_DATA == True :
        log = data_logger.data_log(LOG_DIR, noclobber_logsubdir=False,
                                   log_name="bump_surface")
        log.write_dict_of_arrays(data)
    return data

def bump_save(data, log_dir) :
    "Save the dictionary data, using data_logger.data_log()"
    if log_dir != None :
        log = data_logger.data_log(log_dir, noclobber_logsubdir=False,
                                   log_name="bump")
        log.write_dict_of_arrays(data)

def bump_load(datafile) :
    "Load the dictionary data, using data_logger.date_load()"
    dl = data_logger.data_load()
    data = dl.read_dict_of_arrays(path)
    return data

def bump_plot(data, plotVerbose) :
    "Plot the bump (Vphoto vs Vzp) if plotVerbose or PYLAB_VERBOSE == True"
    if plotVerbose or PYLAB_VERBOSE :
        _import_pylab()
        _pylab.figure(BASE_FIGNUM)
        _pylab.plot(data["Z piezo output"], data["Deflection input"],
                    '.', label='bump')
        _pylab.title("bump surface")
        _pylab.legend(loc='upper left')
        _flush_plot()

def bump(zpiezo, push_depth, npoints=1024, freq=100e3,
         log_dir=None,
         plotVerbose=False) :
    """
    Wrapper around bump_aquire(), bump_save(), bump_analyze(), bump_plot()
    """
    data = bump_aquire(zpiezo, push_depth, npoints, freq)
    bump_save(data, log_dir)
    photoSensitivity = bump_analyze(data, zpiezo.gain, zpiezo.sensitivity,
                                    zpiezo.pos_out2V, zpiezo.def_in2V)
    bump_plot(data, plotVerbose)
    return photoSensitivity

def bump_load_analyze_tweaked(tweak_file, zpGain=_usual_zpGain,
                              zpSensitivity=_usual_zpSensitivity,
                              Vzp_out2V=_usual_Vzp_out2V,
                              Vphoto_in2V=_usual_Vphoto_in2V,
                              plotVerbose=False) :
    "Load the output file of tweak_calib_bump.sh, return an array of slopes"
    photoSensitivity = []
    for line in file(tweak_file, 'r') :
        parsed = line.split()
        path = parsed[0].split('\n')[0]
        # read the data
        full_data = bump_load(path)
        if len(parsed) == 1 :
            data = full_data # use whole bump
        else :
            # use the listed sections
            zp = []
            df = []
            for rng in parsed[1:] :
                p = rng.split(':')
                starti = int(p[0])
                stopi = int(p[1])
                zp.extend(full_data['Z piezo output'][starti:stopi])
                df.extend(full_data['Deflection input'][starti:stopi])
            data = {'Z piezo output': array(zp),
                    'Deflection input':array(df)}
        pSi = bump_analyze(data, zpGain, zpSensitivity,
                           Vzp_out2V, Vphoto_in2V, plotVerbose)
        photoSensitivity.append(pSi)
        bump_plot(data, plotVervose)
    return array(photoSensitivity, dtype=numpy.float)

